==> sagua_0.2.0-1.tar.gz: sagua-0.2.0/src/player_mp3.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3.cc
/// @brief Class for play mp3 source file.

#include <cmath>

#include "../include/mainmanager.h"
#include "../include/player_mp3.h"
#include "../include/trace.h"

MainManager* MAD::MP3Player::mainManager_ = NULL;
Glib::Dispatcher MAD::MP3Player::signalTimerUpdate_;
// struct mad_pcm has samples[2][1152], so is it enough 2304(1152*2) at
// initial size? 
std::vector<double> MAD::MP3Player::bufSample_(2304);

/// @brief Constructor of MP3Player.
///
/// Reset static member variables and connect dispatchers to functions.
MAD::MP3Player::MP3Player(MainManager& mainManager)
{
    mainManager_ = &mainManager;
    
    Reset();
    
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));

}

MAD::MP3Player::~MP3Player()
{
    if(decodeState_() == STOP)
	Clean();
}

/// @brief Initialize MP3Player.
///
/// map file into memory and initialize mad decoder.
/// @param filename filename to play
/// @return Success of initialize
bool MAD::MP3Player::Init(const std::string& filename)
{
    Tracer::Trace trace("MAD::MP3Player::Init");

    filename_() = filename;
    
    if(lstat(filename.c_str(), &buf_.stat) == -1) {
	trace.Debug("Can't get file stat");

	return false;
    }
    
    buf_.fd = open(filename.c_str(), O_RDONLY);
    if(buf_.fd == -1) {
	trace.Debug("Can't open file");

	return false;
    }
    
    void* tempMap = mmap(0, buf_.stat.st_size, PROT_READ, MAP_SHARED, buf_.fd, 0);	
    if(tempMap == MAP_FAILED) {
	trace.Debug("Can't map file");
	
	buf_.fdm = 0;
	buf_.data = 0;
		
	return false;
    } else
    	buf_.fdm = static_cast<unsigned char*>(tempMap);	

    buf_.data = buf_.fdm;

    buf_.length = buf_.stat.st_size;
    buf_.eof = 0;

    mad_decoder_init(&decoder_, &buf_, Input, Header, 0, Output, Error, 0); 

    return true;
}

/// @brief Start player loop
/// 
/// Reset static member variables and start loop. If decodeState is
/// SEEK_REWIND, stop player and seek from begin to wanted position.
///
/// When play is finish, close audio and signal to MainManager.
void MAD::MP3Player::Start()
{
    // reset again cause this is static...
    Reset();
    decodeState_() = PLAY;
    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);
    while(decodeState_() == SEEK_REWIND) {
	Clean();
	timer_().playTimer_ = mad_timer_zero;
	timer_().playFrameCount_ = 0;
	timer_().playTimeSeconds_ = 0;
	Init(filename_());
	decodeState_() = SEEK;	
	mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);
    }

    if(decodeState_() == PLAY)
	mainManager_->AudioClose(true);
    else
	mainManager_->AudioClose(false);
		
    Clean();
    signalFinish_();
}

/// @brief Set decodeState to STOP
///
/// Set decodeState to STOP. Header function detects STOP and send 
/// MAD_FLOW_STOP to decoder.
void MAD::MP3Player::Stop()
{
    if(decodeState_() != STOP)
    	decodeState_() = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds.
///
/// Set decodeState to SEEK. Header function detects SEEK and
/// judge just SEEK or SEEK_REWIND.
/// @param seconds wanted time to seek in seconds.
void MAD::MP3Player::Seek(const unsigned long& seconds)
{
    decodeState_() = SEEK;    

    seekSeconds_() = seconds;
}

/// @brief Reset static member variables.
///
/// Almost member variables are static, need to reset them by hands.
void MAD::MP3Player::Reset()
{
    decodeState_() = STOP;

    timer_().playTimer_ = mad_timer_zero;
    timer_().playFrameCount_ = 0;
    timer_().playTimeSeconds_ = 0;
    
    seekSeconds_() = 0;

}

/// @brief Buffering file to decode.
enum mad_flow MAD::MP3Player::Input(void* data, struct mad_stream* stream)
{
    struct buffer* buf = static_cast<struct buffer*>(data);

    if(buf->eof)
	return MAD_FLOW_STOP;

    if(stream->next_frame) {
	struct stat stat;
	
	if(fstat(buf->fd, &stat) == -1)
	    return MAD_FLOW_BREAK;
	
	unsigned long posn = stream->next_frame - buf->fdm;

	// If you use linux, you get warning about type 
	if(stat.st_size > buf->length) {
	    if(munmap(buf->fdm, buf->length) == -1) {
		buf->fdm = 0;
		buf->data = 0;

		return MAD_FLOW_BREAK;	
	    }

	    buf->length = stat.st_size;
	    buf->fdm = static_cast<unsigned char*>( mmap(0, buf->length, PROT_READ, MAP_SHARED, buf->fd, 0) );

	    if(buf->fdm == 0) {
		buf->data = 0;
		return MAD_FLOW_BREAK;
	    }

	    mad_stream_buffer(stream, buf->fdm + posn, buf->length - posn);

	    return MAD_FLOW_CONTINUE;
	}

	unsigned long left = buf->length - posn;
	
	buf->data = static_cast<unsigned char*>( malloc(left + MAD_BUFFER_GUARD) ); 
	if(buf->data == 0)
	    return MAD_FLOW_BREAK;
	
	buf->eof = 1;
	
	memcpy(buf->data, buf->fdm + posn, left);
	memset(buf->data + left, 0, MAD_BUFFER_GUARD);
	
	mad_stream_buffer(stream, buf->data, left + MAD_BUFFER_GUARD);
	
	return MAD_FLOW_CONTINUE;    
    }

    mad_stream_buffer(stream, buf->fdm, buf->length);
    
    return MAD_FLOW_CONTINUE;
}

/// @brief Output decoded samples.
///
/// Check audio device and open device if need. Put samples to container and
/// send to play function. Update timer and signal to MainManager if timer
/// changed in second.
enum mad_flow MAD::MP3Player::Output(void* data, struct mad_header const* header, struct mad_pcm* pcm)
{
    unsigned int nchannels;
    unsigned int nsamples;
    mad_fixed_t const *left_ch, *right_ch;

    nchannels = pcm->channels;
    nsamples = pcm->length;
    left_ch = pcm->samples[0];
    right_ch = pcm->samples[1];

    if(!mainManager_->AudioCheckOpenDevice() || sampleRate_() != pcm->samplerate) {
	sampleRate_() = pcm->samplerate;
	mainManager_->AudioClose(true);
	mainManager_->AudioOpenDevice(16, nchannels, pcm->samplerate, mainManager_->audioFormatNativeEndian_);
    }
    
    unsigned int index = 0;
    
    for(unsigned int i = 0; i < nsamples; i++)
    {
	index = i * 2;
	bufSample_[index] = *left_ch;
	bufSample_[index] += 0.5;
	bufSample_[index] /= 268435455.5;
	++left_ch;

	index = i * 2 + 1;
	if(nchannels == 2) {
	    bufSample_[index] = *right_ch;
	    bufSample_[index] += 0.5;
	    bufSample_[index] /= 268435455.5;
	    ++right_ch;

	} else
	    bufSample_[index] = 0.0; 

    }

    mainManager_->AudioPlay(bufSample_, nsamples);

    ++timer_().playFrameCount_;
    mad_timer_add(&timer_().playTimer_, header->duration);
    long int currentSeconds = mad_timer_abs(timer_().playTimer_).seconds;

    if( timer_().playTimeSeconds_ != currentSeconds ) {
	
	timer_().playTimeSeconds_ = currentSeconds;
	
	signalTimerUpdate_();

    }
 
    return MAD_FLOW_CONTINUE;
}

/// @brief Control decoder. 
/// @todo Put some cut-off audio output while SEEKING
enum mad_flow MAD::MP3Player::Header(void* data, struct mad_header const* header)
{
    if(decodeState_() == STOP)
	return MAD_FLOW_STOP;
    
    if(decodeState_() == SEEK) {
	if(timer_().playTimer_.seconds < seekSeconds_()) { 
	    mad_timer_add(&timer_().playTimer_, header->duration);
    
	    return MAD_FLOW_IGNORE;
	} else if(timer_().playTimer_.seconds > seekSeconds_()) {
	    decodeState_() = SEEK_REWIND;
	    	
	    return MAD_FLOW_STOP;
	} else {
	    seekSeconds_() = 0;

	    decodeState_() = PLAY;
	}
    }
	    
    return MAD_FLOW_CONTINUE;
}

/// @brief Check decoder's error.
enum mad_flow MAD::MP3Player::Error(void* data, struct mad_stream* stream, struct mad_frame* frame)
{
    Tracer::Trace trace("MAD::MP3Player::Error");
    
    if(stream->error == MAD_ERROR_BADCRC) {
	trace.Debug("MAD_ERROR_BAD_CRC");
		 
	return MAD_FLOW_IGNORE;

    }

    return MAD_FLOW_CONTINUE;
}

/// @brief Create MP3Player object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateMP3Player(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new MAD::MP3Player(mainManager));
}

==> sagua_0.2.1-1.tar.gz: sagua-0.2.1/src/player_mp3.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_mp3.cc
/// @brief Class for play mp3 source file.

#include <cmath>

#include <string>

#include "../include/mainmanager.h"
#include "../include/player_mp3.h"
#include "../include/trace.h"

MainManager* MAD::MP3Player::mainManager_ = NULL;
Glib::Dispatcher MAD::MP3Player::signalTimerUpdate_;
// struct mad_pcm has samples[2][1152], so is it enough 2304(1152*2) at
// initial size? 
std::vector<double> MAD::MP3Player::bufSample_(2304);
std::string MAD::MP3Player::filename_;
DecodeState MAD::MP3Player::decodeState_;
unsigned int MAD::MP3Player::sampleRate_;
MAD::Timer MAD::MP3Player::timer_;
long MAD::MP3Player::seekSeconds_;

/// @brief Constructor of MP3Player.
///
/// Reset static member variables and connect dispatchers to functions.
MAD::MP3Player::MP3Player(MainManager& mainManager)
{
    mainManager_ = &mainManager;
    
    Reset();
    
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
    signalTimerUpdate_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayTimeUpdate));

}

MAD::MP3Player::~MP3Player()
{
    if(decodeState_ == STOP)
	Clean();
}

/// @brief Initialize MP3Player.
///
/// map file into memory and initialize mad decoder.
/// @param filename filename to play
/// @return Success of initialize
bool MAD::MP3Player::Init(const std::string& filename)
{
    Tracer::Trace trace("MAD::MP3Player::Init");

    filename_ = filename;
    
    if(lstat(filename.c_str(), &buf_.stat) == -1) {
	trace.Debug("Can't get file stat");

	return false;
    }
    
    buf_.fd = open(filename.c_str(), O_RDONLY);
    if(buf_.fd == -1) {
	trace.Debug("Can't open file");

	return false;
    }
    
    void* tempMap = mmap(0, buf_.stat.st_size, PROT_READ, MAP_SHARED, buf_.fd, 0);	
    if(tempMap == MAP_FAILED) {
	trace.Debug("Can't map file");
	
	buf_.fdm = 0;
	buf_.data = 0;
		
	return false;
    } else
    	buf_.fdm = static_cast<unsigned char*>(tempMap);	

    buf_.data = buf_.fdm;

    buf_.length = buf_.stat.st_size;
    buf_.eof = 0;

    mad_decoder_init(&decoder_, &buf_, Input, Header, 0, Output, Error, 0); 

    return true;
}

/// @brief Start player loop
/// 
/// Reset static member variables and start loop. If decodeState is
/// SEEK_REWIND, stop player and seek from begin to wanted position.
///
/// When play is finish, close audio and signal to MainManager.
void MAD::MP3Player::Start()
{
    // reset again cause this is static...
    Reset();
    decodeState_ = PLAY;
    mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);
    while(decodeState_ == SEEK_REWIND) {
	Clean();
	timer_.playTimer_ = mad_timer_zero;
	timer_.playFrameCount_ = 0;
	timer_.playTimeSeconds_ = 0;
	Init(filename_);
	decodeState_ = SEEK;	
	mad_decoder_run(&decoder_, MAD_DECODER_MODE_SYNC);
    }

    if(decodeState_ == PLAY)
	mainManager_->AudioClose(true);
    else
	mainManager_->AudioClose(false);
		
    Clean();
    signalFinish_();
}

/// @brief Set decodeState to STOP
///
/// Set decodeState to STOP. Header function detects STOP and send 
/// MAD_FLOW_STOP to decoder.
void MAD::MP3Player::Stop()
{
    if(decodeState_ != STOP)
    	decodeState_ = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds.
///
/// Set decodeState to SEEK. Header function detects SEEK and
/// judge just SEEK or SEEK_REWIND.
/// @param seconds wanted time to seek in seconds.
void MAD::MP3Player::Seek(const unsigned long& seconds)
{
    decodeState_ = SEEK;    

    seekSeconds_ = seconds;
}

/// @brief Reset static member variables.
///
/// Almost member variables are static, need to reset them by hands.
void MAD::MP3Player::Reset()
{
    decodeState_ = STOP;

    timer_.playTimer_ = mad_timer_zero;
    timer_.playFrameCount_ = 0;
    timer_.playTimeSeconds_ = 0;
    
    seekSeconds_ = 0;

}

/// @brief Buffering file to decode.
enum mad_flow MAD::MP3Player::Input(void* data, struct mad_stream* stream)
{
    struct buffer* buf = static_cast<struct buffer*>(data);

    if(buf->eof)
	return MAD_FLOW_STOP;

    if(stream->next_frame) {
	struct stat stat;
	
	if(fstat(buf->fd, &stat) == -1)
	    return MAD_FLOW_BREAK;
	
	const unsigned long posn = stream->next_frame - buf->fdm;

	// If you use linux, you get warning about type 
	if(stat.st_size > buf->length) {
	    if(munmap(buf->fdm, buf->length) == -1) {
		buf->fdm = 0;
		buf->data = 0;

		return MAD_FLOW_BREAK;	
	    }

	    buf->length = stat.st_size;
	    buf->fdm = static_cast<unsigned char*>( mmap(0, buf->length, PROT_READ, MAP_SHARED, buf->fd, 0) );

	    if(buf->fdm == 0) {
		buf->data = 0;
		return MAD_FLOW_BREAK;
	    }

	    mad_stream_buffer(stream, buf->fdm + posn, buf->length - posn);

	    return MAD_FLOW_CONTINUE;
	}

	const unsigned long left = buf->length - posn;
	
	buf->data = static_cast<unsigned char*>( malloc(left + MAD_BUFFER_GUARD) ); 
	if(buf->data == 0)
	    return MAD_FLOW_BREAK;
	
	buf->eof = 1;
	
	memcpy(buf->data, buf->fdm + posn, left);
	memset(buf->data + left, 0, MAD_BUFFER_GUARD);
	
	mad_stream_buffer(stream, buf->data, left + MAD_BUFFER_GUARD);
	
	return MAD_FLOW_CONTINUE;    
    }

    mad_stream_buffer(stream, buf->fdm, buf->length);
    
    return MAD_FLOW_CONTINUE;
}

/// @brief Output decoded samples.
///
/// Check audio device and open device if need. Put samples to container and
/// send to play function. Update timer and signal to MainManager if timer
/// changed in second.
enum mad_flow MAD::MP3Player::Output(void* data, struct mad_header const* header, struct mad_pcm* pcm)
{
    const unsigned int nchannels = pcm->channels;
    const unsigned int nsamples = pcm->length;
    const mad_fixed_t * left_ch = pcm->samples[0];
    const mad_fixed_t * right_ch = pcm->samples[1];

    if(!mainManager_->AudioCheckOpenDevice() || sampleRate_ != pcm->samplerate) {
	sampleRate_ = pcm->samplerate;
	mainManager_->AudioClose(true);
	mainManager_->AudioOpenDevice(16, nchannels, pcm->samplerate, mainManager_->audioFormatNativeEndian_);
    }
    
    // ugly but for little more speed
    unsigned int index = 0;
    if(nchannels == 2) {
	
	for(unsigned int i = 0; i < nsamples; ++i) {
	    
	    bufSample_[index] = *left_ch;
	    bufSample_[index] += 0.5;
	    bufSample_[index] /= 268435455.5;
	    ++index;
	    ++left_ch;

	    bufSample_[index] = *right_ch;
	    bufSample_[index] += 0.5;
	    bufSample_[index] /= 268435455.5;
	    ++index;
	    ++right_ch;
	}
	
    } else {

	for(unsigned int i = 0; i < nsamples; ++i) {
	   	    
	    bufSample_[index] = *left_ch;
	    bufSample_[index] += 0.5;
	    bufSample_[index] /= 268435455.5;
	    ++index;
	    ++left_ch;

	    bufSample_[index] = 0.0;
	    ++index;

	}

    }

    mainManager_->AudioPlay(bufSample_, nsamples);

    ++timer_.playFrameCount_;
    mad_timer_add(&timer_.playTimer_, header->duration);
    const long int currentSeconds = mad_timer_abs(timer_.playTimer_).seconds;

    if( timer_.playTimeSeconds_ != currentSeconds ) {
	
	timer_.playTimeSeconds_ = currentSeconds;
	
	signalTimerUpdate_();

    }
 
    return MAD_FLOW_CONTINUE;
}

/// @brief Control decoder. 
/// @todo Put some cut-off audio output while SEEKING
enum mad_flow MAD::MP3Player::Header(void* data, struct mad_header const* header)
{
    if(decodeState_ == STOP)
	return MAD_FLOW_STOP;
    
    if(decodeState_ == SEEK) {
	if(timer_.playTimer_.seconds < seekSeconds_) { 
	    mad_timer_add(&timer_.playTimer_, header->duration);
    
	    return MAD_FLOW_IGNORE;
	} else if(timer_.playTimer_.seconds > seekSeconds_) {
	    decodeState_ = SEEK_REWIND;
	    	
	    return MAD_FLOW_STOP;
	} else {
	    seekSeconds_ = 0;

	    decodeState_ = PLAY;
	}
    }
	    
    return MAD_FLOW_CONTINUE;
}

/// @brief Check decoder's error.
enum mad_flow MAD::MP3Player::Error(void* data, struct mad_stream* stream, struct mad_frame* frame)
{
    Tracer::Trace trace("MAD::MP3Player::Error");
    
    if(stream->error == MAD_ERROR_BADCRC) {
	trace.Debug("MAD_ERROR_BAD_CRC");
		 
	return MAD_FLOW_IGNORE;

    }

    return MAD_FLOW_CONTINUE;
}

/// @brief Create MP3Player object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateMP3Player(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new MAD::MP3Player(mainManager));
}
